# Written by Amanda Moody, Central Washington University, Ellensburg, WA. 2019. Provided freely "as is."
# RUSLE L Factor calculation according to the RUSLE Handbook (Renard et al. 1997) and following the Grid Cumulation (GC)
# method for calculating slope length originally proposed by Hickey et al. (1994) for USLE.
# Original RUSLE based AML code published by Van Remortel et al. (2001). Major revisions include no longer using
# ESRI focal flow tool to define high points and updating the non-cumulative slope length (NCSL) calculations for
# high points. S Factor calculation uses the Max. Downhill Slope method (Hickey 2000).


# It is highly recommended to run code with a 64-bit download of python and a solid state hard drive.
# If unsure open python command line and read the top line of code to determine if you are using 32 or 64-bit python.

### Navigation ### approx. lines
# Maximum Downhill Slope calculation (MDS) - LINE 114
# S Factor calculation - LINE 191
# L Factor components - line 222
    # NCSL - LINE 223
    # CDHSL - LINE 333
    # L Factor Calculation - LINE 539

import arcpy, numpy, math, time, sys
arcpy.AddMessage("Beginning process at " + time.ctime())

myRaster = arcpy.GetParameterAsText(0) # input DEM
outputL = arcpy.GetParameterAsText(1) # output location for L factor raster
outputS = arcpy.GetParameterAsText(2) # output location for S factor raster
outputSlope = arcpy.GetParameterAsText(3) # output location for slope raster (no slope will equal 0 degrees)
# slope cutoff angles for slopes less than 5% and those greater or equal to 5%
less5 = float(arcpy.GetParameterAsText(4)) # if unsure, recommended value .5
great5 = float(arcpy.GetParameterAsText(5)) # if unsure, recommended value .5
# this is the percent of max flow accumulation value to designate defined channels (where RUSLE is not applicable)
# rule of thumb for stream threshold is 1% while 100% would nullify this variable
threshold = float(arcpy.GetParameterAsText(6)) # if unsure, .5 for more defined channels, 1 for general stream channels,
# or 100 if you don't want this to influence results

# custom raise statement for inappropriate inputs by the user
class InputError(Exception):
    pass

if less5 < 0 or less5 > 1:
    raise InputError("Invalid slope cutoff value. Please input a value between 0 and 1.")

if great5 < 0 or great5 > 1:
    raise InputError("Invalid slope cutoff value. Please input a value between 0 and 1.")

if threshold < 0 or threshold > 100:
    raise InputError("Invalid defined channel threshold value. Please input a value between 0 and 100.")

# get working raster's info. to use when turning l factor array into a raster
desc = arcpy.Describe(myRaster)
sr = desc.spatialReference
lLeft = desc.extent.lowerLeft
cSize = arcpy.Raster(myRaster).meanCellHeight
units = sr.linearUnitName

# units must be either feet or meters, output L factor will be in the same units as DEM. while RUSLE guidelines use
# feet, users continue L factor calculation with meters as that is the international system of measurement for length.
if "Feet" in units or "Foot" in units or "Meter" in units:
    pass
else:
    raise InputError("Invalid linear unit. Please input a DEM using feet or meters.")

### Step: Define Functions ###
# buffer edge cells to nodata (can't know true value at edge cells, missing surrounding spatial info.) for any function
# that uses surrounding cells to calculate center cell value
def noDataBuff(noArray, topLeft, bottomRight):
    noArray[:, topLeft] = -9999 # first column starts at 0
    noArray[topLeft, :] = -9999 # first row starts 0
    noArray[:, (nCols - bottomRight)] = -9999 # last column in array is the # of columns - 1 (0 indexed)
    noArray[(nRows - bottomRight), :] = -9999 # last row in array is the # of row - 1 (0 indexed)

# locate and define new row and col for a specific neighboring cell. example, neighbor cell above origin cell has
# coordinates originRow-1, originCol. if no change for row/col use 0. used in "Step: Calc NCSL" and "Step: Calc CDHSL"
def nbrCell(originRow, originCol, nbrRow, nbrCol):
    varR = originRow + nbrRow
    varC = originCol + nbrCol
    if varR < 0 or varR == nRows:
        return "Invalid"
    elif varC < 0 or varC == nCols:
        return "Invalid"
    else:
        return varR, varC

# calc. a cell's S Factor value, input cell slope in degrees and percent
def sCalc(slope, percent):
    if percent < 9:
        sValue = 10.8 * (math.sin(math.radians(slope))) + 0.03
    elif percent >= 9:
        sValue = 16.8 * (math.sin(math.radians(slope))) - 0.50
    return sValue

# calc. a cell's L Factor value, input the slope angle, length, and if unit plot is 22.13 (m) or 72.6 (ft)
def Lcalc(slopeVar, lengthVar, unit):
    beta = (math.sin(math.radians(slopeVar))/0.0896)/(3*((math.sin(math.radians(slopeVar)))**0.8)+0.56)
    m = (beta/(1+beta))
    if "Foot" in units or "Feet" in units:
        lValue = (lengthVar/72.6)**m
    elif "Meter" in units:
        lValue = (lengthVar/22.13)**m
    return lValue

# calculate slope percent used in "Step: Calc S Factor" and "Step: Calc CDHSL"
def slopePerc(value):
    return (math.tan(math.radians(value))*100)

arcpy.CheckOutExtension("Spatial")
### Step: fill DEM ###
fillRaster = arcpy.sa.Fill(myRaster)

### Step: Calc flow direction using D8 algorithm (ESRI 2019 Flow Direction) ###
arcpy.AddMessage("Getting flow direction.")
flowDirc = arcpy.sa.FlowDirection(fillRaster, "FORCE")
# force edge cells to flow outwards to keep edge effect errors consistent across site
arcpy.CheckInExtension("Spatial")

### Step: Calc max. downhill slope angle (Dunn and Hickey 1998) ###
# Original code written by Dr. Sterling Quinn with some logic adapted from code posted by user FelixIP on StackExchange
# https://gis.stackexchange.com/questions/136715/getting-cell-value-along-flow-direction-using-arcpy
arcpy.AddMessage("Calculating maximum downhill slope angle.")

fDirc = (1, 2, 4, 8, 16, 32, 64, 128)
rookDirc = (1, 4, 16, 64)
diagDirc = (2, 8, 32, 128)
# fdCol and fdRow are used from indexed fDirc value to correctly reference cell in the flow direction
fdCol = (1, 1, 0, -1, -1, -1, 0, 1)
fdRow = (0, 1, 1, 1, 0, -1, -1, -1)

fDircArray = arcpy.RasterToNumPyArray(flowDirc, "", "", "", -9999)

fAccD8 = arcpy.sa.FlowAccumulation(flowDirc) # for defined channel threshold
del flowDirc, myRaster

# this is the array shape of original DEM, used for any new array being built
nRows, nCols = fDircArray.shape
cTotal = nRows * nCols

elevArray = arcpy.RasterToNumPyArray(fillRaster, "", "", "", -9999)
del fillRaster

maxDHSArray = numpy.empty((nRows, nCols))

for nRow in range(nRows):
    for nCol in range(nCols):
        # find elevation and flow direction of current cell at location nRow, nCol
        elevPixel = elevArray[nRow, nCol]
        if elevPixel == -9999:# nodata value
            maxDHSArray[nRow, nCol] = -9999
            continue

        fDircPixel = fDircArray[nRow, nCol]
        if fDircPixel == -9999:
            maxDHSArray[nRow, nCol] = -9999
            continue
        # this indicates the pixel in the direction of fDircPixel's flow
        i = fDirc.index(fDircPixel)
        # get location of the comparing cell (the cell in the flow direction)
        newRow = nRow + fdRow[i]
        # this accounts for those instances where
        # the flow direction points to a non-existent row outside raster coverage
        if newRow < 0 or newRow == nRows:
            maxDHSArray[nRow, nCol] = -9999
            continue
        newCol = nCol + fdCol[i]
        if newCol < 0 or newCol == nCols:
            maxDHSArray[nRow, nCol] = -9999
            continue

        # now the elevation of that comparing cell can be referenced
        newElevPixel = elevArray[newRow, newCol]
        # calculate the difference
        elevDiff = elevPixel - newElevPixel

        # calculate max downhill slope for current cell
        # if comparing cell in diagonal direction then divide by orthogonal size
        if fDircPixel in diagDirc:
            maxDHSArray[nRow, nCol] = math.degrees(math.atan(float(elevDiff)/(1.4142*cSize)))
        # if comparing cell in cardinal direction then divide by cell size
        elif fDircPixel in rookDirc:
            maxDHSArray[nRow, nCol] = math.degrees(math.atan(float(elevDiff)/cSize))
        else:
            maxDHSArray[nRow, nCol] = -9999

if outputSlope and outputSlope != "#":
    maxDHSraster = arcpy.NumPyArrayToRaster(maxDHSArray, lLeft, cSize, cSize, -9999)
    arcpy.DefineProjection_management(maxDHSraster, sr)
    maxDHSraster.save(outputSlope)
    del maxDHSraster
    
# value = 0 change to 0.1. this allows for erosion in every cell without altering flow paths (following the GC method)
maxDHSArray[maxDHSArray == 0] = 0.1
noDataBuff(maxDHSArray, 0, 1)

### Step: Calc S Factor ### following RUSLE guidelines (Agricultural Handbook No. 703)
# use the maximum downhill slope angle calculated previously because the neighborhood method (ESRI 2019) averages
# slope from all neighbors and smooths the landscape and local variability
faccD8Array = arcpy.RasterToNumPyArray(fAccD8, "", "", "", -9999)
del fAccD8
flowAccD8Max = numpy.nanmax(faccD8Array)
streamsD8 = float(threshold / 100) * flowAccD8Max

if outputS and outputS != "#":
    arcpy.AddMessage("Calculating S Factor.")

    sFactorArray = numpy.empty((nRows, nCols))

    for nRow in range(nRows):
        for nCol in range(nCols):
            slope = maxDHSArray[nRow, nCol]
            slopeP = slopePerc((maxDHSArray[nRow, nCol]))
            faccPixel = faccD8Array[nRow, nCol]

            if slope == -9999:
                sFactorArray[nRow, nCol] = -9999
            elif faccPixel > streamsD8:  # if location is designated as a "defined" channel, RUSLE does not apply
                sFactorArray[nRow, nCol] = -9999
            else:
                sFactorArray[nRow, nCol] = sCalc(slope, slopeP)

    SFactor = arcpy.NumPyArrayToRaster(sFactorArray, lLeft, cSize, cSize, -9999)
    arcpy.DefineProjection_management(SFactor, sr)
    SFactor.save(outputS)
    del sFactorArray, SFactor

### Step: L Factor Calc Components (GC Method)###
## Step: Calc NCSL ##
# NCSL for each cell calculated by flow direction and designation as a high point or flat area
arcpy.AddMessage("Calculating non-cumulative slope length.")
NCSLArray = numpy.empty((nRows, nCols))

# loop through rasters high points and flow direction
for nRow in range(nRows):
    for nCol in range(nCols):

        fDircPixel = fDircArray[nRow, nCol]

        # set row and column locations for all cell neighbors around current cell, this helps determines those cells
        # that have no inflow from their surrounding neighbors, according to the D8 algorithm
        tCell = nbrCell(nRow, nCol, -1, 0)
        if tCell == "Invalid":
            tFD = -9999
        else:
            tFD = fDircArray[tCell]

        rCell = nbrCell(nRow, nCol, 0, 1)
        if rCell == "Invalid":
            rFD = -9999
        else:
            rFD = fDircArray[rCell]

        bCell = nbrCell(nRow, nCol, 1, 0)
        if bCell == "Invalid":
            bFD = -9999
        else:
            bFD = fDircArray[bCell]

        lCell = nbrCell(nRow, nCol, 0, -1)
        if lCell == "Invalid":
            lFD = -9999
        else:
            lFD = fDircArray[lCell]

        trCell = nbrCell(nRow, nCol, -1, 1)
        if trCell == "Invalid":
            trFD = -9999
        else:
            trFD = fDircArray[trCell]

        tlCell = nbrCell(nRow, nCol, -1, -1)
        if tlCell == "Invalid":
            tlFD = -9999
        else:
            tlFD = fDircArray[tlCell]

        brCell = nbrCell(nRow, nCol, 1, 1)
        if brCell == "Invalid":
            brFD = -9999
        else:
            brFD = fDircArray[brCell]

        blCell = nbrCell(nRow, nCol, 1, -1)
        if blCell == "Invalid":
            blFD = -9999
        else:
            blFD = fDircArray[blCell]

        # apply NCSL rules to write NSCLArray values
        # if cell high point (no flow into it) then multiply by .5 to account for only length downhill from center
        if fDircPixel == -9999:
            NCSLArray[nRow, nCol] = -9999
        elif tFD != 4 and trFD != 8 and rFD != 16 and brFD != 32 and bFD != 64 and blFD != 128 and lFD != 1 and tlFD != 2:
            if fDircPixel in rookDirc:
                NCSLArray[nRow, nCol] = .5 * cSize
            elif fDircPixel in diagDirc:
                NCSLArray[nRow, nCol] = .5 * 1.4142 * cSize
        # if receiving cell in cardinal direction then set equal to cell size
        elif fDircPixel in rookDirc:
            NCSLArray[nRow, nCol] = cSize
        # if receiving cell in diagonal direction than multiply cell size by orthogonal distance
        elif fDircPixel in diagDirc:
            NCSLArray[nRow, nCol] = 1.4142 * cSize
        else:
            NCSLArray[nRow, nCol] = -9999

# loop through again to check for flat areas that should be changed to .5 that value
for nRow in range(nRows):
    for nCol in range(nCols):

        elevPixel = elevArray[nRow, nCol]
        fDircPixel = fDircArray[nRow, nCol]
        if fDircPixel == -9999:
            continue
        i = fDirc.index(fDircPixel)

        newRow = nRow + fdRow[i]
        if newRow < 0 or newRow == nRows:
            continue
        newCol = nCol + fdCol[i]
        if newCol < 0 or newCol == nCols:
            continue

        newElevPixel = elevArray[newRow, newCol]

        # if cell's elevation is equal to receiving cell's elevation, then receiving cell is a flat area
        if elevPixel == newElevPixel:
            if fDircArray[newRow, newCol] in rookDirc:
                NCSLArray[newRow, newCol] = .5 * cSize
            elif fDircArray[newRow, newCol] in diagDirc:
                NCSLArray[newRow, newCol] = .5 * 1.4142 * cSize
        else:
            continue

noDataBuff(NCSLArray, 0, 1)
del elevArray

## Step: Calc cumulative downhill slope length (CDHSL) ##
try:
    CDHSLArray = numpy.zeros((nRows, nCols))

    # identify high points and no data values in CHDSL, these are the points that will start the cumulation process
    for nRow in range(nRows):
        for nCol in range(nCols):

            NCSLPixel = NCSLArray[nRow, nCol]

            if NCSLPixel == (.5 * cSize) or NCSLPixel == (.5 * 1.4142 * cSize):
                CDHSLArray[nRow, nCol] = NCSLArray[nRow, nCol]
            elif NCSLPixel == -9999:
                CDHSLArray[nRow, nCol] = -9999
            else:
                continue

    # iterate through array adding NCSL from the high points/flat areas down along flow direction. since any cell that
    # does not have flow into it is defined as a high/flat area and given a value above, all remaining cells should
    # receive some length value (no cell has 0 degree slope so always some slope length). this can take up to 12 hours.
    arcpy.AddMessage("Beginning cumulative downhill slope calculations at " + time.ctime() + ". This may take a while.")
    computron = 0
    while 0 in CDHSLArray:
        computron += 1
        if (computron/50.0).is_integer():
            arcpy.AddMessage("Calculating...")
        for nRow in range(nRows):
            for nCol in range(nCols):

                CDHSLPixel = CDHSLArray[nRow, nCol]
                # must start where a CDHSL value exists, this means first iteration at high points
                if CDHSLPixel == -9999 or CDHSLPixel == 0:
                    continue

                fDircPixel = fDircArray[nRow, nCol]
                i = fDirc.index(fDircPixel)
                newRow = nRow + fdRow[i]
                if newRow < 0 or newRow == nRows:
                    continue
                newCol = nCol + fdCol[i]
                if newCol < 0 or newCol == nCols:
                    continue

                NCSLPixel = NCSLArray[nRow, nCol]
                newNCSL = NCSLArray[newRow, newCol]
                newCDHSL = CDHSLArray[newRow, newCol]
                if newCDHSL != 0: # if receiving cell has CDHSL then all in flows already considered and calculated
                    # for that cell. No need to repeat lengthy calculation process over again.
                    continue

                if maxDHSArray[newRow, newCol] == -9999: # slope value must exist for L factor and slope cutoff consideration
                    CDHSLArray[newRow, newCol] = -9999
                    continue
                else:
                    newSlope = math.tan(math.radians(maxDHSArray[newRow, newCol]))*100
                # set rows & columns locations for all neighbors around the receiving cell (cell in the flow direction)
                tCell = nbrCell(newRow, newCol, -1, 0)
                if tCell == "Invalid":
                    tFD = -9999
                    tCDHSL = -9999
                    tSlope = -9999
                else:
                    tFD = fDircArray[tCell]
                    tCDHSL = CDHSLArray[tCell]
                    tSlope = slopePerc((maxDHSArray[tCell]))

                rCell = nbrCell(newRow, newCol, 0, 1)
                if rCell == "Invalid":
                    rFD = -9999
                    rCDHSL = -9999
                    rSlope = -9999
                else:
                    rFD = fDircArray[rCell]
                    rCDHSL = CDHSLArray[rCell]
                    rSlope = slopePerc((maxDHSArray[rCell]))

                bCell = nbrCell(newRow, newCol, 1, 0)
                if bCell == "Invalid":
                    bFD = -9999
                    bCDHSL = -9999
                    bSlope = -9999
                else:
                    bFD = fDircArray[bCell]
                    bCDHSL = CDHSLArray[bCell]
                    bSlope = slopePerc((maxDHSArray[bCell]))

                lCell = nbrCell(newRow, newCol, 0, -1)
                if lCell == "Invalid":
                    lFD = -9999
                    lCDHSL = -9999
                    lSlope = -9999
                else:
                    lFD = fDircArray[lCell]
                    lCDHSL = CDHSLArray[lCell]
                    lSlope = slopePerc((maxDHSArray[lCell]))

                trCell = nbrCell(newRow, newCol, -1, 1)
                if trCell == "Invalid":
                    trFD = -9999
                    trCDHSL = -9999
                    trSlope = -9999
                else:
                    trFD = fDircArray[trCell]
                    trCDHSL = CDHSLArray[trCell]
                    trSlope = slopePerc((maxDHSArray[trCell]))

                tlCell = nbrCell(newRow, newCol, -1, -1)
                if tlCell == "Invalid":
                    tlFD = -9999
                    tlCDHSL = -9999
                    tlSlope = -9999
                else:
                    tlFD = fDircArray[tlCell]
                    tlCDHSL = CDHSLArray[tlCell]
                    tlSlope = slopePerc((maxDHSArray[tlCell]))

                brCell = nbrCell(newRow, newCol, 1, 1)
                if brCell == "Invalid":
                    brFD = -9999
                    brCDHSL = -9999
                    brSlope = -9999
                else:
                    brFD = fDircArray[brCell]
                    brCDHSL = CDHSLArray[brCell]
                    brSlope = slopePerc((maxDHSArray[brCell]))

                blCell = nbrCell(newRow, newCol, 1, -1)
                if blCell == "Invalid":
                    blFD = -9999
                    blCDHSL = -9999
                    blSlope = -9999
                else:
                    blFD = fDircArray[blCell]
                    blCDHSL = CDHSLArray[blCell]
                    blSlope = slopePerc((maxDHSArray[blCell]))

                # if empty receiving cell then cont. check other neighbors if they also flow to the same receiving cell
                pNeighbors = []
                slopes = []
                if tFD == 4 and tCDHSL != -9999 and tSlope != -9999:
                    pNeighbors.append(tCDHSL)
                    slopes.append(tSlope)
                if trFD == 8 and trCDHSL != -9999 and trSlope != -9999:
                    pNeighbors.append(trCDHSL)
                    slopes.append(trSlope)
                if rFD == 16 and rCDHSL != -9999 and rSlope != -9999:
                    pNeighbors.append(rCDHSL)
                    slopes.append(rSlope)
                if brFD == 32 and brCDHSL != -9999 and brSlope != -9999:
                    pNeighbors.append(brCDHSL)
                    slopes.append(brSlope)
                if bFD == 64 and bCDHSL != -9999 and bSlope != -9999:
                    pNeighbors.append(bCDHSL)
                    slopes.append(bSlope)
                if blFD == 128 and blCDHSL != -9999 and blSlope != -9999:
                    pNeighbors.append(blCDHSL)
                    slopes.append(blSlope)
                if lFD == 1 and lCDHSL != -9999 and lSlope != -9999:
                    pNeighbors.append(lCDHSL)
                    slopes.append(lSlope)
                if tlFD == 2 and tlCDHSL != -9999 and tlSlope != -9999:
                    pNeighbors.append(tlCDHSL)
                    slopes.append(tlSlope)

                if 0 in pNeighbors: # don't need to calculate slope changes if one of the neighbors CDHSL is still a zero
                    continue
                else:
                    # neighbors who are eligible to pass length. slope change doesn't trigger their slope to reset
                    finalNeighbors = []
                    # check if slope cutoff is triggered for any neighbors, they will not be eligible to pass length
                    if newSlope >= 5:
                        slopeIndex = 0
                        for neighbor in pNeighbors: # starts at neighbor index 0 which corresponds to start slopeIndex 0
                            if ((slopes[slopeIndex] - newSlope) / (slopes[slopeIndex]) * 100) >= (great5*100):
                                slopeIndex += 1
                            # if slope cutoff not triggered, CDHSL neighbor able to pass on length
                            else:
                                finalNeighbors.append(neighbor)
                                slopeIndex += 1
                        del slopeIndex
                    elif newSlope < 5:
                        slopeIndex = 0
                        for neighbor in pNeighbors:
                            if ((slopes[slopeIndex] - newSlope) / (slopes[slopeIndex]) * 100) >= (less5*100):
                                slopeIndex += 1
                            else:
                                finalNeighbors.append(neighbor)
                                slopeIndex += 1
                        del slopeIndex
                    del slopes, pNeighbors

                    # determine cell's CDHSL value
                    if len(finalNeighbors) == 0: # slope cutoff met for all neighbors and cell resets slope length
                        CDHSLArray[newRow, newCol] = newNCSL
                    else: # from neighbors list, use max value (greatest slope length) to add for new CDHSL
                        CDHSLArray[newRow, newCol] = max(finalNeighbors) + newNCSL
                    del finalNeighbors
except:
    message = sys.exc_info()[1]
    arcpy.AddError(message.args[0])

del fDircArray, NCSLArray
arcpy.AddMessage("Completed cumulative downhill slope calculations at " + time.ctime() + ".")

noDataBuff(CDHSLArray, 1, 2)

## Step: L factor calculation ## following RUSLE guidelines (Agricultural Handbook No. 703)
arcpy.AddMessage("Calculating L Factor.")

try:
    lFactorArray = numpy.empty((nRows, nCols))

    for nRow in range(nRows):
        for nCol in range(nCols):
            slope = maxDHSArray[nRow, nCol]
            length = CDHSLArray[nRow, nCol]
            faccPixel = faccD8Array[nRow, nCol]

            if slope == -9999 or length == -9999:
                lFactorArray[nRow, nCol] = -9999
            elif faccPixel > streamsD8:  # if location is designated as a "defined" channel, RUSLE does not apply
                lFactorArray[nRow, nCol] = -9999
            else:
                lFactorArray[nRow, nCol] = Lcalc(slope, length, units)

    del CDHSLArray, faccD8Array
    LFactor = arcpy.NumPyArrayToRaster(lFactorArray, lLeft, cSize, cSize, -9999)
    arcpy.DefineProjection_management(LFactor, sr)
    LFactor.save(outputL)
    del lFactorArray, LFactor
except:
    message = sys.exc_info()[1]
    arcpy.AddError(message.args[0])
